import os

ATASKQ_CONNECTION = os.getenv("ATASKQ_CONNECTION", "sqlite://ataskq.db.sqlite3")
ATASKQ_TASK_WAIT_TIMEOUT = os.getenv("ATASKQ_TASK_WAIT_TIMEOUT", None) and float(os.environ["ATASKQ_TASK_WAIT_TIMEOUT"])
ATASKQ_TASK_PULL_INTERVAL = float(os.getenv("ATASKQ_TASK_PULL_INTERVAL", 0.2))
ATASKQ_MONITOR_PULSE_INTERVAL = float(os.getenv("ATASKQ_MONITOR_PULSE_INTERVAL", 60))
ATASKQ_TASK_PULSE_TIMEOUT = float(os.getenv("ATASKQ_TASK_PULSE_TIMEOUT", 60 * 5))

ATASKQ_SERVER_CONNECTION = os.getenv("ATASKQ_SERVER_CONNECTION", "sqlite://ataskq.db.sqlite3")
ATASKQ_SERVER_TASK_PULSE_TIMEOUT_MONITOR_INTERVAL = float(
    os.getenv("ATASKQ_SERVER_TASK_PULSE_TIMEOUT_MONITOR_INTERVAL", 60 * 5)
)

ATASKQ_DB_LIMIT_DEFAULT = int(os.getenv("ATASKQ_DB_LIMIT_DEFAULT", 100))
ATASKQ_DB_INIT_ON_HANDLER_INIT = bool(os.getenv("ATASKQ_DB_INIT_ON_HANDLER_INIT", True))
